/*
 *  TOPPERS/SSP Kernel
 *      Smallest Set Profile Kernel
 *
 *  Copyright (C) 2013-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2018,2019 by Naoki Saito
 *             Nagoya Municipal Industrial Research Institute, JAPAN
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  $Id: core_timer.c 386 2019-04-17 03:13:00Z nmir-saito $
 */

/*
 *  タイマドライバ（Generic Timer）
 */
#include "kernel_impl.h"
#include "time_event.h"
#include "target_timer.h"

/*
 *  タイマの周波数(単位: kHz)
 */
static uint32_t timer_clock;

/*
 *  タイマの設定値(比較対象とする値)
 */
static uint64_t timer_cval;

/*
 *  タイマの開始
 */
Inline void
target_timer_start(void)
{
    uint32_t cntp_ctl;

    cntp_ctl = clock_read_ctrl();
    cntp_ctl |= ((uint32_t)CNTP_CTL_ENABLE);
    clock_write_ctrl(cntp_ctl);
}

/*
 *  タイマの停止
 */
Inline void
target_timer_stop(void)
{
    uint32_t cntp_ctl;
    cntp_ctl = clock_read_ctrl();
    cntp_ctl &= ~CNTP_CTL_ENABLE;
    clock_write_ctrl(cntp_ctl);
}

/*
 *  タイマの起動処理
 */
void
target_timer_initialize(intptr_t exinf)
{
    uint32_t cyc;
    uint32_t timer_clock_hz;

    /*
     *  タイマの停止
     */
    target_timer_stop();

    /*
     *  プリスケーラ(=2^31 --> divider=1)
     */
    sil_wrw_mem((uint32_t *)(0x40000008), 0x80000000);

    timer_clock_hz = get_timer_freq();   // 19.2MHz
    timer_clock = timer_clock_hz / 1000;

    /*
     *  1ティックのクロック数
     */
    cyc = (timer_clock * (TIC_NUME) / (TIC_DENO));

    /*
     *  タイマの開始
     */
    clock_write_tval((uint32_t)cyc);
    target_timer_start();

    /*
     *  タイマの設定値
     */
    timer_cval = clock_read_cval();
}

/*
 *  タイマの停止処理
 */
void
target_timer_terminate(intptr_t exinf)
{
    /*
     *  タイマの停止
     */
    clock_write_ctrl((uint32_t)0);
}

/*
 *  タイマ割込みハンドラ
 */
void
target_timer_handler(void)
{
    uint32_t cyc;
    uint64_t pct;

    /*
     *  タイマの停止
     */
    target_timer_stop();

    /*
     *  1ティックのクロック数
     */
    cyc = (timer_clock * (TIC_NUME) / (TIC_DENO));

    /*
     *  タイマの設定値
     */
    timer_cval = timer_cval + (uint64_t)cyc;


    // カウンタから読み出し
    pct = clock_read_counter();

    while (timer_cval > pct + cyc) {
        timer_cval = timer_cval - (uint64_t)cyc;
    }
    while (timer_cval < pct) {
        timer_cval = timer_cval + (uint64_t)cyc;
    }

    /*
     *  タイマの設定
     */
    clock_write_cval((uint64_t)timer_cval);

    /*
     *  タイマの開始
     */
    target_timer_start();

    i_begin_int(INTNO_TIMER);
    signal_time();              /* タイムティックの供給 */
    i_end_int(INTNO_TIMER);
}
